############################################
# System dependencies
############################################
import numpy as np

import netin
from . import constants as const
from .sampling import Sampling


############################################
# Class
############################################
class RandomNodes(Sampling):
    """Random node sampling.

    Parameters
    ----------
    g: netin.Graph | netin.DiGraph
        global network

    pseeds: float
        fraction of seeds to sample

    max_tries: int
        maximum number of tries to sample a subgraph with enough classes and edges

    random_seed: object
        seed for random number generator

    kwargs: dict
        additional parameters for the sampling method
    """

    ######################################################
    # Constructor
    ######################################################
    def __init__(self, g: netin.Graph, pseeds: float, max_tries: int = const.MAX_TRIES,
                 random_seed: object = None, **kwargs):
        super().__init__(g, pseeds, max_tries, random_seed, **kwargs)

    @property
    def method_name(self) -> str:
        return const.RANDOM_NODES

    def sampling(self):
        super().sampling()

    def _sample(self):
        """
        Creates a subgraph from G based on random node sampling
        """
        num_classes = 0
        edges = None

        ### 1. pick random nodes
        while num_classes < const.MIN_CLASSES:
            nodes = list(self.g.nodes())
            np.random.shuffle(nodes)
            nodes = nodes[:self.nseeds]
            num_classes = self._count_classes(nodes)

        return nodes, edges



from netin import PATC


def run():
    n = 200
    k = 2
    f_m = 0.1
    tc = 0.9
    seed = 1234
    g = PATC(n=n, k=k, f_m=f_m, tc=tc, seed=seed)
    g.generate()
    g.info()


if __name__ == '__main__':
    run()



############################################
# System dependencies
############################################
from typing import List
import networkx as nx
import numpy as np
import gc

############################################
# Local dependencies
############################################
import netin
from netin.utils import validator as val
from netin.stats import networks as net
from . import constants as const


############################################
# Class
############################################
class Sampling(object):
    """Base class for sampling methods.

    Parameters
    ----------
    g: netin.Graph | netin.DiGraph
        global network

    pseeds: float
        fraction of seeds to sample

    max_tries: int
        maximum number of tries to sample a subgraph with enough classes and edges

    random_seed: object
        seed for random number generator

    kwargs: dict
        additional parameters for the sampling method

    Notes
    -----
    - The original graph ``g`` (passed as parameter) is not modified.
       The sampling method creates a copy of it, and stores it in ``self.g``.
    - This class does not create a subgraph.

    """

    def __init__(self, g: netin.Graph, pseeds: float, max_tries: int = const.MAX_TRIES,
                 random_seed: object = None, **kwargs):
        self.g = g.copy()
        self.pseeds = pseeds
        self.max_tries = max_tries
        self.random_seed = random_seed
        self.nseeds = int(pseeds * self.g.number_of_nodes())
        self.sample = None
        self.nodes = None
        self.train_index = None
        self.test_index = None
        self.membership_y = None
        self.feature_x = None
        self.test_nodes = None
        self.kwargs = kwargs
        np.random.seed(self.random_seed)

    def sampling(self):
        """
        Creates a new instance of the respective sampling method, and calls its respective extract_subgraph method.
        """
        val.validate_float(self.pseeds, 0, 1)
        val.validate_more_than_one(net.get_node_attributes(self.g))
        self.sample = nx.DiGraph() if self.g.is_directed() else nx.Graph()
        self._extract_subgraph()
        self._set_graph_metadata()

    @property
    def method_name(self) -> str:
        """
        Name of sampling method.
        """
        return ''

    def _count_classes(self, nodes: List) -> int:
        """
        Counts the number of classes in a given set of nodes

        Parameters
        ----------
        nodes: list
            list of nodes

        Returns
        -------
        int
            number of classes
        """
        return len(set([self.g.nodes[n][self.g.graph['class_attribute']] for n in nodes]))

    def _extract_subgraph(self):
        """
        Creates a subgraph from G based on the sampling technique
        """
        max_tries = self.max_tries or const.MAX_TRIES
        num_edges = 0
        tries = -1

        while num_edges < const.MIN_EDGES:

            tries += 1
            if tries >= max_tries:
                raise RuntimeWarning(f"The sample has not enough edges ({num_edges}), and max_tries has exceeded. "
                                     "Try increasing the number of tries or increasing the number of seeds.")
                return

            nodes, edges = self._sample()

            ### 2. recreate induced subgraph
            sample = self.g.copy()

            # removing edges
            if edges:
                edges_to_remove = set(sample.edges()) - edges
                sample.remove_edges_from(edges_to_remove)

            # removing nodes
            if nodes:
                nodes_to_remove = [n for n in self.g.nodes() if n not in nodes]
                sample.remove_nodes_from(nodes_to_remove)

            num_edges = sample.number_of_edges()

        if num_edges < const.MIN_EDGES:
            raise RuntimeWarning("The sample has no edges.")

        self.sample = sample.copy()
        self.sample.node_list = [n for n in self.sample.node_list if n in self.sample.nodes()]
        self.sample.node_class_values = {n: l for n, l in self.sample.node_class_values.items() if n in self.sample.nodes()}
        gc.collect()

    def _set_graph_metadata(self):
        """
        Updates the training sample subgraph metadata
        """
        self.sample.graph['method'] = self.method_name
        self.sample.graph['pseeds'] = self.pseeds
        nx.set_node_attributes(G=self.g, name='seed', values={n: int(n in self.sample) for n in self.g.nodes()})
        self.sample.graph['m'] = net.get_min_degree(self.sample)
        self.sample.graph['d'] = nx.density(self.sample)
        self.sample.graph['n'] = self.sample.number_of_nodes()
        self.sample.graph['f_m'] = net.get_minority_fraction(self.sample)
        self.sample.graph['similarity'] = net.get_similitude(self.sample)
        self.sample.graph['e'] = self.sample.number_of_edges()
        k, km, kM = net.get_average_degrees(self.sample)
        self.sample.graph['k'] = k
        self.sample.graph['km'] = km
        self.sample.graph['kM'] = kM
        self.sample.graph['random_seed'] = self.random_seed
        self.sample.graph['original_graph'] = self.sample.graph['model']
        del (self.sample.graph['model'])
        self.sample.set_model_name(f"{self.sample.get_model_name()}\n{self.method_name}")

        # for LINK: working with matrices
        self.nodes = list(self.g.nodes())
        self.train_index = np.array([i for i, n in enumerate(self.nodes) if n in self.sample])
        self.test_nodes, self.test_index = zip(*[(n, i) for i, n in enumerate(self.nodes) if n not in self.sample])
        self.test_index = np.array(self.test_index)
        self.feature_x = nx.adjacency_matrix(self.g, self.nodes).toarray()
        self.membership_y = np.array(
            [self.g.graph['class_values'].index(self.g.nodes[n][self.g.graph['class_attribute']]) for n in self.nodes])

    def info(self):
        """
        Prints a summary of the training sample subgraph, including its attributes.

        """
        print(nx.info(self.sample))
        print(self.sample.graph)



from typing import Union
from typing import Tuple
import warnings

import numpy as np
import networkx as nx
from collections import Counter

from netin.utils import constants as const
from netin.utils import validator as val


def get_min_degree(g: Union[nx.Graph, nx.DiGraph]) -> int:
    """
    Returns the minimum degree of nodes in the graph.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to compute the minimum degree

    Returns
    -------
    int
        Minimum degree of nodes in the graph
    """
    degrees = [d for n, d in g.degree]
    return min(degrees) if len(degrees) > 0 else -1


def get_minority_fraction(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> float:
    """
    Computes the fraction of the minority class in the graph.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to compute the fraction of the minority class

    """
    n = g.number_of_nodes()
    majority, minority, class_attribute = _get_class_labels(g, class_attribute)
    minority_count = sum([1 for n, obj in g.nodes(data=True) if obj[class_attribute] == minority])
    f_m = minority_count / n

    return f_m


def get_edge_type_counts(g: Union[nx.Graph, nx.DiGraph], fractions: bool = False,
                         class_attribute: str = None) -> Counter:
    """
    Computes the edge type counts of the graph using the `class_attribute` of each node.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to compute the edge type counts

    fractions: bool
        If True, the counts are returned as fractions of the total number of edges.

    class_attribute: str
        The name of the attribute that holds the class label of each node.

    Returns
    -------
    Counter
        Counter holding the edge type counts

    Notes
    -----
    Class labels are assumed to be binary. The minority class is assumed to be labeled as 1.
    """
    majority, minority, class_attribute = _get_class_labels(g, class_attribute)
    class_values = [majority, minority]
    class_labels = [const.MAJORITY_LABEL, const.MINORITY_LABEL]

    counts = Counter([f"{class_labels[class_values.index(g.nodes[e[0]][class_attribute])]}"
                      f"{class_labels[class_values.index(g.nodes[e[1]][class_attribute])]}"
                      for e in g.edges if g.nodes[e[0]][class_attribute] in class_values and
                      g.nodes[e[1]][class_attribute] in class_values])

    if fractions:
        total = sum(counts.values())
        counts = Counter({k: v / total for k, v in counts.items()})

    return counts


def get_average_degree(g: Union[nx.Graph, nx.DiGraph]) -> float:
    """
    Returns the average node degree of the graph.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to compute the average degree for

    Returns
    -------
    float
        Average degree of the graph
    """
    k = sum([d for n, d in g.degree]) / g.number_of_nodes()
    return k


def get_average_degrees(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> Tuple[float, float, float]:
    """
    Computes and returns the average degree of the graph, the average degree of the majority and the minority class.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to compute the average degree for

    class_attribute: str
        Name of the class attribute in the graph

    Returns
    -------
    Tuple[float, float, float]
        Average degree of the graph, the average degree of the majority and the minority class
    """
    k = get_average_degree(g)

    majority, minority, class_attribute = _get_class_labels(g, class_attribute)
    kM = np.mean([d for n, d in g.degree if g.nodes[n][class_attribute] == majority])
    km = np.mean([d for n, d in g.degree if g.nodes[n][class_attribute] == minority])

    return k, kM, km


def get_similitude(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> float:
    """
    Computes and returns the fraction of same-class edges in the graph.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to compute the similitude for

    class_attribute: str
        Name of the class attribute in the graph

    Returns
    -------
    float
        Fraction of same-class edges in the graph
    """
    majority, minority, class_attribute = _get_class_labels(g, class_attribute)

    tmp = [int(g.nodes[e[0]][class_attribute] == g.nodes[e[1]][class_attribute]) for e in g.edges]
    total = len(tmp)
    sim = sum(tmp) / total

    return sim


def get_node_attributes(g: Union[nx.Graph, nx.DiGraph]) -> list:
    """
    Returns the values of the class attribute for all nodes in the graph.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to get the node attributes from

    Returns
    -------
    list
        List of node attributes
    """
    val.validate_graph_metadata(g)
    l = [a for n, a in nx.get_node_attributes(g, g.graph['class_attribute']).items()]
    return l


def _get_graph_metadata_value(g: Union[nx.Graph, nx.DiGraph], key: str, default: object = None) -> Union[object, iter]:
    value = default if key not in g.graph or g.graph[key] is None else g.graph[key]
    return value


def _get_class_labels(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> Tuple[str, str, str]:
    if class_attribute:
        counter = Counter([obj[class_attribute] for n, obj in g.nodes(data=True)])
    else:
        val.validate_graph_metadata(g)
        class_attribute = _get_graph_metadata_value(g, 'class_attribute', const.CLASS_ATTRIBUTE)
        counter = Counter([obj[class_attribute] for n, obj in g.nodes(data=True)])

    if len(counter) > 2:
        warnings.warn(f'Graph contains more than two classes: {counter}')

    majority = counter.most_common()[0][0]
    minority = counter.most_common()[1][0]

    return majority, minority, class_attribute



from typing import Set
from typing import Tuple
from typing import Union

import networkx as nx
import numpy as np
import powerlaw

from netin.utils import constants as const
from netin.utils import validator as val
from .graph import Graph


class UnDiGraph(Graph):
    """Undirected graph base model.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    k: int
        minimum degree of nodes (minimum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is an undirected graph with n nodes and no edges.
    Then, everytime a node is selected as source, it gets connected to k target nodes.
    Target nodes are selected depending on the chosen mechanism of edge formation:

    - PA: Preferential attachment (in-degree), see :class:`netin.PA` [BarabasiAlbert1999]_
    - PAH: Preferential attachment (in-degree) with homophily, see :class:`netin.PAH` [Karimi2018]_
    - PATC: Preferential attachment (in-degree) with triadic closure, see :class:`netin.PATC` [HolmeKim2002]_
    - PATCH: Preferential attachment (in-degree) with homophily and triadic closure, see :class:`netin.PATCH`

    References
    ----------
    .. [BarabasiAlbert1999] A. L. Barabasi and R. Albert "Emergence of scaling in random networks", Science 286, pp 509-512, 1999.
    .. [Karimi2018] F. Karimi, M. Génois, C. Wagner, P. Singer, & M. Strohmaier, M "Homophily influences ranking of minorities in social networks", Scientific reports 8(1), 11077, 2018.
    .. [HolmeKim2002] P. Holme and B. J. Kim “Growing scale-free networks with tunable clustering” Phys. Rev. E 2002.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, k: int, f_m: float, seed: object = None):
        Graph.__init__(self, n=n, f_m=f_m, seed=seed)
        self.k = k

    ############################################################
    # Init
    ############################################################

    def _validate_parameters(self):
        """
        Validates the parameters of the undirected.
        """
        super()._validate_parameters()
        val.validate_int(self.k, minimum=1)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns metadata for a undirected.
        """
        obj = super().get_metadata_as_dict()
        obj.update({
            'k': self.k,
        })
        return obj

    ############################################################
    # Generation
    ############################################################

    def get_target(self, source: Union[None, int], targets: Union[None, Set[int]],
                   special_targets: Union[None, object, iter]) -> int:
        """
        Picks a random target node based on preferential attachment.

        Parameters
        ----------
        special_targets : object
            Special targets to be considered

        source: int
            Newly added node

        targets: Set[int]
            Potential target nodes in the undirected based on preferential attachment

        Returns
        -------
            int
                Target node that an edge should be added to
        """
        # Collect probabilities to connect to each node in target_list
        target_set = set([t for t in targets if t != source and t not in nx.neighbors(self, source)])
        probs, target_set = self.get_target_probabilities(source, target_set, special_targets)
        return np.random.choice(a=list(target_set), size=1, replace=False, p=probs)[0]

    def generate(self):
        """
        An undirected graph of n nodes is grown by attaching new nodes each with k edges.
        Each edge is either drawn by preferential attachment, homophily, and/or triadic closure.

        For triadic closure, a candidate is chosen uniformly at random from all triad-closing edges (of the new node).
        Otherwise, or if there are no triads to close, edges are connected via preferential attachment and/or homophily.

        Homophily varies ranges from 0 (heterophilic) to 1 (homophilic), where 0.5 is neutral.
        Similarly, triadic closure varies from 0 (no triadic closure) to 1 (full triadic closure).

        - PA: An undirected graph with h_mm = h_MM in [0.5, None] and tc = 0 is a BA preferential attachment model.
        - PAH: An undirected graph with h_mm not in [0.5, None] and h_MM not in [0.5, None] and tc = 0 is a PA model with homophily.
        - PATC: An undirected graph with h_mm = h_MM in [0.5, None] and tc > 0 is a PA model with triadic closure.
        - PATCH: An undirected graph with h_mm not in [0.5, None] and h_MM not in [0.5, None] and tc > 0 is a PA model with homophily and triadic closure.

        """
        # 1. Init an undirected graph and nodes (assign class labels)
        super().generate()

        # 2. Iterate until n nodes are added (starts with k pre-existing, unconnected nodes)
        for source in self.node_list[self.k:]:
            targets = set(range(source))  # targets via preferential attachment
            special_targets = self.get_special_targets(source)

            for idx_target in range(self.k):
                # Choose next target
                target = self.get_target(source, targets, special_targets)

                special_targets = self.update_special_targets(idx_target, source, target, targets, special_targets)

                # Finally add edge to undirected
                self.add_edge(source, target)

                # Call event handlers if present
                self.on_edge_added(source, target)

        self._terminate()

    ############################################################
    # Getters and Setters
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.UNDIRECTED_MODEL_NAME)

    def get_expected_number_of_edges(self) -> int:
        """
        Computes and returns the expected number of edges based on minimum degree `k` and number of nodes `n`

        Returns
        -------
        int
            Expected number of edges
        """
        return (self.get_expected_number_of_nodes() * self.get_expected_minimum_degree()) - \
            (self.get_expected_minimum_degree() ** self.get_expected_minimum_degree())

    def get_expected_minimum_degree(self) -> int:
        """
        Returns the expected minimum degree of the graph (`k`, the input parameter)

        Returns
        -------
        int
            Expected minimum degree
        """

        return self.k

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the parameters of the model.
        """
        print('k: {}'.format(self.k))

    def info_computed(self):
        """
        Shows the computed properties of the graph.
        """
        fit_M, fit_m = self.fit_degree_powerlaw()
        print(f"- Powerlaw fit (degree):")
        print(f"- {self.get_majority_label()}: alpha={fit_M.power_law.alpha}, "
              f"sigma={fit_M.power_law.sigma}, "
              f"min={fit_M.power_law.xmin}, max={fit_M.power_law.xmax}")
        print(f"- {self.get_minority_label()}: alpha={fit_m.power_law.alpha}, "
              f"sigma={fit_m.power_law.sigma}, "
              f"min={fit_m.power_law.xmin}, max={fit_m.power_law.xmax}")

    def fit_degree_powerlaw(self) -> Tuple[powerlaw.Fit, powerlaw.Fit]:
        """
        Returns the powerlaw fit of the degree distribution to a powerlaw for the majority and minority class.

        Returns
        -------
        fit_M : powerlaw.Fit
            Powerlaw fit for the majority class

        fit_m: powerlaw.Fit
            Powerlaw fit for the minority class
        """
        vM = self.get_majority_value()
        dM = [d for n, d in self.degree() if self.nodes[n][self.class_attribute] == vM]
        dm = [d for n, d in self.degree() if self.nodes[n][self.class_attribute] != vM]

        fit_M = powerlaw.Fit(data=dM, discrete=True, xmin=min(dM), xmax=max(dM))
        fit_m = powerlaw.Fit(data=dm, discrete=True, xmin=min(dm), xmax=max(dm))
        return fit_M, fit_m

    def calculate_degree_powerlaw_exponents(self) -> Tuple[float, float]:
        """
        Returns the powerlaw exponents for the majority and minority class.

        Returns
        -------
        pl_M : float
            Powerlaw exponent for the majority class

        pl_m: float
            Powerlaw exponent for the minority class
        """
        pl_M, pl_m = calculate_degree_powerlaw_exponents(self)

        # fit_M, fit_m = self.fit_degree_powerlaw()
        # pl_M = fit_M.power_law.alpha
        # pl_m = fit_m.power_law.alpha

        return pl_M, pl_m

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        return self.__class__(n=self.n,
                              k=self.k,
                              f_m=self.f_m,
                              seed=self.seed)


def calculate_degree_powerlaw_exponents(g: UnDiGraph) -> Tuple[float, float]:
    """
    Returns the power law exponents for the in-degree distribution of the majority and minority class.

    Parameters
    ----------
    g: DiGraph
        Graph to calculate the power law exponents for

    Returns
    -------
    Tuple[float, float]
        power law exponents for the in-degree distribution of the majority and minority class
    """
    fit_M, fit_m = g.fit_powerlaw(metric='degree')
    pl_M = fit_M.power_law.alpha
    pl_m = fit_m.power_law.alpha
    return pl_M, pl_m



from .dh import DH
from .directed import DiGraph
from .directed import calculate_out_degree_powerlaw_exponents
from .directed import calculate_in_degree_powerlaw_exponents
from .dpa import DPA
from .dpah import DPAH
from .graph import Graph
from .graph import convert_networkx_to_netin
from .h import Homophily
from .pa import PA
from .pah import PAH
from .patc import PATC
from .patch import PATCH
from .tc import TriadicClosure
from .undirected import UnDiGraph
from .undirected import calculate_degree_powerlaw_exponents



from typing import Union, Set, Tuple

import numpy as np
from sympy import Eq
from sympy import solve
from sympy import symbols

import netin
from netin.generators.h import Homophily
from netin.utils import constants as const
from .pa import PA


class PAH(PA, Homophily):
    """Creates a new PAH instance. An undirected graph with preferential attachment and homophily.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    k: int
        minimum degree of nodes (minimum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    h_MM: float
        homophily (similarity) between majority nodes (minimum=0, maximum=1.)

    h_mm: float
        homophily (similarity) between minority nodes (minimum=0, maximum=1.)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is an undirected with n nodes, where f_m are the minority.
    Then, everytime a node is selected as source, it gets connected to k target nodes.
    Target nodes are selected via preferential attachment (in-degree) and homophily (h_**).
    This model is based on [Karimi2018]_ known as the "Barabasi model with homophily" or "BA Homophily".
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, k: int, f_m: float, h_MM: float, h_mm: float, seed: object = None):
        PA.__init__(self, n=n, k=k, f_m=f_m, seed=seed)
        Homophily.__init__(self, n=n, f_m=f_m, h_MM=h_MM, h_mm=h_mm, seed=seed)

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.PAH_MODEL_NAME)

    def _validate_parameters(self):
        """
        Validates the parameters of the undirected.
        """
        PA._validate_parameters(self)
        Homophily._validate_parameters(self)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns the metadata (parameters) of the model as a dictionary.

        Returns
        -------
        dict
            metadata of the model
        """
        obj = PA.get_metadata_as_dict(self)
        obj.update(Homophily.get_metadata_as_dict(self))
        return obj

    ############################################################
    # Generation
    ############################################################

    def _initialize(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the model.

        Parameters
        ----------
        class_attribute: str
            name of the attribute that represents the class

        class_values: list
            values of the class attribute

        class_labels: list
            labels of the class attribute mapping the class_values.
        """
        PA._initialize(self, class_attribute, class_values, class_labels)
        Homophily._initialize(self, class_attribute, class_values, class_labels)

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                 special_targets: Union[None, object, iter] = None) -> tuple[np.array, set[int]]:
        """
        Returns the probabilities of selecting a target node from a set of nodes based on the preferential attachment
        and homophily.

        Parameters
        ----------
        source: int
            source node

        target_set: set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        tuple[np.array, set[int]]
            probabilities of selecting a target node from a set of nodes, and the set of target nodes
        """
        probs = np.array([self.get_homophily_between_source_and_target(source, target) *
                          (self.degree(target) + const.EPSILON) for target in target_set])
        probs /= probs.sum()
        return probs, target_set

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the parameters of the model.
        """
        PA.info_params(self)
        Homophily.info_params(self)

    def info_computed(self):
        """
        Shows the computed properties of the graph.
        """
        PA.info_computed(self)
        Homophily.info_computed(self)

    def infer_homophily_values(self) -> Tuple[float, float]:
        """
        Infers the level of homophily using the analytical solution of the model.

        Returns
        -------
        tuple[float, float]
            homophily between majority nodes, and homophily between minority nodes

        Notes
        -----
        See derivations in [Karimi2018]_.
        """

        h_MM, h_mm = infer_homophily(self)

        return h_MM, h_mm

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        return self.__class__(n=self.n,
                              k=self.k,
                              f_m=self.f_m,
                              h_MM=self.h_MM,
                              h_mm=self.h_mm,
                              seed=self.seed)

    @staticmethod
    def fit(g, n=None, k=None, seed=None):
        """
        It fits the PAH model to the given graph.

        Parameters
        ----------
        g: netin.UnDiGraph
            graph to fit the model to

        n: int
            number of nodes to override (e.g., to generate a smaller network)

        k: int
            minimum node degree to override (e.g., to generate a denser network ``k>1``)

        seed: object
            seed for random number generator

        Returns
        -------
        netin.PAH
            fitted model
        """
        n = n or g.number_of_nodes()
        k = k or g.calculate_minimum_degree()
        f_m = g.calculate_fraction_of_minority()
        h_MM, h_mm = infer_homophily(g)

        new_g = PAH(n=n,
                    k=k,
                    f_m=f_m,
                    h_MM=float(h_MM),
                    h_mm=float(h_mm),
                    seed=seed)
        new_g.generate()

        return new_g


def infer_homophily(g) -> Tuple[float, float]:
    f_m = g.calculate_fraction_of_minority()
    f_M = 1 - f_m

    e = g.calculate_edge_type_counts()
    e_MM = e['MM']
    e_mm = e['mm']
    M = e['MM'] + e['mm'] + e['Mm'] + e['mM']

    p_MM = e_MM / M
    p_mm = e_mm / M

    pl_M, pl_m = g.calculate_degree_powerlaw_exponents()
    b_M = -1 / (pl_M + 1)
    b_m = -1 / (pl_m + 1)

    # equations
    hmm, hMM, hmM, hMm = symbols('hmm hMM hmM hMm')
    eq1 = Eq((f_m * f_m * hmm * (1 - b_M)) / ((f_m * hmm * (1 - b_M)) + (f_M * hmM * (1 - b_m))), p_mm)
    eq2 = Eq(hmm + hmM, 1)

    eq3 = Eq((f_M * f_M * hMM * (1 - b_m)) / ((f_M * hMM * (1 - b_m)) + (f_m * hMm * (1 - b_M))), p_MM)
    eq4 = Eq(hMM + hMm, 1)

    solution = solve((eq1, eq2, eq3, eq4), (hmm, hmM, hMM, hMm))
    h_MM, h_mm = solution[hMM], solution[hmm]
    return h_MM, h_mm



from collections import defaultdict
from typing import Union, Set, Tuple

import networkx as nx
import numpy as np
import powerlaw

from netin.utils import constants as const
from netin.utils import validator as val
from .graph import Graph


class DiGraph(nx.DiGraph, Graph):
    """Directed graph base model.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    d: float
        edge density (minimum=0, maximum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    plo_M: float
        activity (out-degree power law exponent) majority group (minimum=1)

    plo_m: float
        activity (out-degree power law exponent) minority group (minimum=1)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is a directed with n nodes and no edges.
    Source nodes are selected based on their activity given by plo_M (if majority) or plo_m (if minority).
    Target nodes are selected depending on the chosen mechanism of edge formation.

    - DPAH: preferential attachment (in-degree) and homophily (h**), see :class:`netin.DPAH`
    - DPA: preferential attachment (in-degree), see :class:`netin.DPA`
    - DH: homophily (h**), see :class:`netin.DH`

    References
    ----------
    .. [Espin-Noboa2022] L. Espín-Noboa, C. Wagner, M. Strohmaier, & F. Karimi "Inequality and inequity in network-based ranking and recommendation algorithms" Scientific reports 12(1), 1-14, 2022.
    .. [Karimi2018] F. Karimi, M. Génois, C. Wagner, P. Singer, & M. Strohmaier, M "Homophily influences ranking of minorities in social networks", Scientific reports 8(1), 11077, 2018.
    .. [BarabasiAlbert1999] A. L. Barabasi and R. Albert "Emergence of scaling in random networks", Science 286, pp 509-512, 1999.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, f_m: float, d: float, plo_M: float, plo_m: float, seed: object = None):
        nx.DiGraph.__init__(self)
        Graph.__init__(self, n=n, f_m=f_m, seed=seed)
        self.d = d
        self.plo_M = plo_M
        self.plo_m = plo_m
        self.in_degrees = None
        self.out_degrees = None
        self.activity = None
        self.expected_number_of_edges = None

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.DIRECTED_MODEL_NAME)

    def _validate_parameters(self):
        """
        Validates the parameters of the directed.
        """
        Graph._validate_parameters(self)
        val.validate_float(self.d, minimum=1. / (self.n * (self.n - 1)), maximum=1.)
        val.validate_float(self.plo_M, minimum=1. + const.EPSILON)
        val.validate_float(self.plo_m, minimum=1. + const.EPSILON)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns metadata for a directed.
        """
        obj = super().get_metadata_as_dict()
        obj.update({
            'd': self.d,
            'plo_M': self.plo_M,
            'plo_m': self.plo_m,
        })
        return obj

    ############################################################
    # Generation
    ############################################################

    def _initialize(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the model.

        Parameters
        ----------
        class_attribute: str
            name of the attribute that represents the class

        class_values: list
            values of the class attribute

        class_labels: list
            labels of the class attribute mapping the class_values.
        """
        Graph._initialize(self, class_attribute, class_values, class_labels)
        self._init_edges()
        self._init_activity()

    def _init_edges(self):
        """
        Initializes the expected number of edges based on the number of nodes and density of the graph (input param).
        It also initializes the in- and out-degrees of the nodes.
        """
        self.expected_number_of_edges = int(round(self.d * self.n * (self.n - 1)))
        self.in_degrees = np.zeros(self.n)
        self.out_degrees = np.zeros(self.n)

    def _init_activity(self):
        """
        Initializes the level of activity for each node based on the power law exponents (input param).
        """
        act_M = powerlaw.Power_Law(parameters=[self.plo_M], discrete=True).generate_random(self.n_M)
        act_m = powerlaw.Power_Law(parameters=[self.plo_m], discrete=True).generate_random(self.n_m)
        self.activity = np.append(act_M, act_m)
        if np.inf in self.activity:
            self.activity[self.activity == np.inf] = 0.0
            self.activity += 1
        self.activity /= self.activity.sum()

    def get_sources(self) -> np.array:
        """
        Returns a random sample with replacement of nodes to be used as source nodes.
        The sample has the length of the expected number of edges, and the probability of each node to be selected is
        based on its activity.

        Returns
        -------
        np.array
            array of source nodes
        """
        return np.random.choice(a=np.arange(self.n), size=self.expected_number_of_edges, replace=True, p=self.activity)

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int], np.array],
                                 special_targets: Union[None, object, iter] = None) -> np.array:
        pass

    def get_target(self, source: int, edge_list: dict, **kwargs) -> Union[None, int]:
        """
        Returns a target node for a given source node.

        Parameters
        ----------
        source: int
            source node

        edge_list: dict
            dictionary of edges

        kwargs: dict
            additional parameters

        Returns
        -------
        Union[None, int]
            target node
        """
        one_percent = self.n * 1 / 100.
        if np.count_nonzero(self.out_degrees) > one_percent:
            targets = [n for n in np.arange(self.n) if n not in edge_list[source]]
        else:
            targets = np.arange(self.n)
        targets = np.delete(targets, np.where(targets == source))

        if targets.shape[0] == 0:
            return None

        probs = self.get_target_probabilities(source, targets, **kwargs)
        return np.random.choice(a=targets, size=1, replace=False, p=probs)[0]

    def generate(self):
        """
        A directed graph of n nodes is grown by attaching new nodes.
        Source nodes are selected randomly with replacement based on their activity.
        Each target node drawn based on the chosen mechanism of edge formation.

        - DPA: A graph with h_mm = h_MM in [0.5, None] is a directed BA preferential attachment model, see :class:`netin.DPA`.
        - DH: A graph with h_mm not in [0.5, None] and h_MM not in [0.5, None] is a directed Erdos-Renyi with homophily, see :class:`netin.DPH`.
        - DPAH: A graph with h_mm not in [0.5, None] and h_MM not in [0.5, None] is a DPA model with homophily, see :class:`netin.DPAH`.
        """
        # 1. Init directed and nodes (assign class labels)
        Graph.generate(self)

        # 2. Iterate until reaching desired number of edges (edge density)
        tries = 0
        edge_list = defaultdict(list)
        while self.number_of_edges() < self.expected_number_of_edges:
            tries += 1
            for source in self.get_sources():
                target = self.get_target(source, edge_list)

                if target is None:
                    continue

                if not self.has_edge(source, target):
                    self.add_edge(source, target)
                    self.in_degrees[target] += 1
                    self.out_degrees[source] += 1
                    edge_list[source].append(target)

                if self.number_of_edges() >= self.expected_number_of_edges:
                    break

            # if no more edges can be added, break
            if tries > const.MAX_TRIES_EDGE and self.number_of_edges() < self.expected_number_of_edges:
                print(f">> Edge density ({nx.density(self)}) might differ from {self.d:.5f} (n={self.n}, f_m={self.f_m}"
                      f"seed={self.seed}, plo_M={self.plo_M}, plo_m={self.plo_m}")
                break

        self._terminate()

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the (input) parameters of the graph.
        """
        print(f'd: {self.d} (expected edges: {self.expected_number_of_edges})')
        print(f'plo_M: {self.plo_M}')
        print(f'plo_m: {self.plo_m}')

    def info_computed(self):
        """
        Shows the computer properties of the graph.
        """
        for metric in ['in_degree', 'out_degree']:
            fit_M, fit_m = self.fit_powerlaw(metric)
            print(f"- Powerlaw fit ({metric}):")
            print(f"- {self.get_majority_label()}: alpha={fit_M.power_law.alpha}, sigma={fit_M.power_law.sigma}, "
                  f"min={fit_M.power_law.xmin}, max={fit_M.power_law.xmax}")
            print(f"- {self.get_minority_label()}: alpha={fit_m.power_law.alpha}, sigma={fit_m.power_law.sigma}, "
                  f"min={fit_m.power_law.xmin}, max={fit_m.power_law.xmax}")

    def calculate_in_degree_powerlaw_exponents(self) -> Tuple[float, float]:
        """
        Returns the power law exponents for the in-degree distribution of the majority and minority class.

        Returns
        -------
        Tuple[float, float]
            power law exponents for the in-degree distribution of the majority and minority class
        """
        pl_M, pl_m = calculate_in_degree_powerlaw_exponents(self)

        # fit_M, fit_m = self.fit_powerlaw(metric='in_degree')
        # pl_M = fit_M.power_law.alpha
        # pl_m = fit_m.power_law.alpha

        return pl_M, pl_m

    def calculate_out_degree_powerlaw_exponents(self) -> Tuple[float, float]:
        """
        Returns the power law exponents for the out-degree distribution of the majority and minority class.

        Returns
        -------
        Tuple[float, float]
            power law exponents for the out-degree distribution of the majority and minority class
        """
        pl_M, pl_m = calculate_out_degree_powerlaw_exponents(self)

        # fit_M, fit_m = self.fit_powerlaw(metric='out_degree')
        # pl_M = fit_M.power_law.alpha
        # pl_m = fit_m.power_law.alpha

        return pl_M, pl_m

    ############################################################
    # Getters and setters
    ############################################################

    def get_expected_number_of_edges(self) -> int:
        """
        Returns the expected number of edges based on number of nodes and edge density.

        Returns
        -------
        int
            expected number of edges
        """
        return self.expected_number_of_edges

    def get_expected_density(self) -> float:
        """
        Returns the expected edge density (d, the input parameter).

        Returns
        -------
        float
            expected edge density
        """
        return self.d

    def get_expected_powerlaw_out_degree_majority(self) -> float:
        """
        Returns the expected power law exponent for the out-degree distribution of the majority class
        (plo_M, the input parameter).

        Returns
        -------
        float
            expected power law exponent for the out-degree distribution of the majority class
        """
        return self.plo_M

    def get_expected_powerlaw_out_degree_minority(self):
        """
        Returns the expected power law exponent for the out-degree distribution of the minority class
        (plo_m, the input parameter).

        Returns
        -------
        float
            expected power law exponent for the out-degree distribution of the minority class
        """
        return self.plo_m

    def get_activity_distribution(self) -> np.array:
        """
        Returns the activity distribution of all the nodes in the graph.

        Returns
        -------
        np.array
            activity distribution of all the nodes in the graph
        """
        return self.activity

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        obj = self.__class__(n=self.n,
                             d=self.d,
                             f_m=self.f_m,
                             plo_M=self.plo_M,
                             plo_m=self.plo_m,
                             seed=self.seed)
        return obj


def calculate_in_degree_powerlaw_exponents(g: DiGraph) -> Tuple[float, float]:
    """
    Returns the power law exponents for the in-degree distribution of the majority and minority class.

    Parameters
    ----------
    g: DiGraph
        Graph to calculate the power law exponents for

    Returns
    -------
    Tuple[float, float]
        power law exponents for the in-degree distribution of the majority and minority class
    """
    fit_M, fit_m = g.fit_powerlaw(metric='in_degree')
    pl_M = fit_M.power_law.alpha
    pl_m = fit_m.power_law.alpha
    return pl_M, pl_m


def calculate_out_degree_powerlaw_exponents(g: DiGraph) -> Tuple[float, float]:
    """
    Returns the power law exponents for the out-degree distribution of the majority and minority class.

    Parameters
    ----------
    g: DiGraph
        Graph to calculate the power law exponents for

    Returns
    -------
    Tuple[float, float]
        power law exponents for the out-degree distribution of the majority and minority class
    """
    fit_M, fit_m = g.fit_powerlaw(metric='out_degree')
    pl_M = fit_M.power_law.alpha
    pl_m = fit_m.power_law.alpha
    return pl_M, pl_m



from netin import PATCH


def run():
    n = 200
    k = 2
    f_m = 0.1
    h_MM = 0.1
    h_mm = 0.9
    tc = 0.5
    seed = 1234
    g = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, tc=tc, seed=seed)
    g.generate()
    g.info()


if __name__ == '__main__':
    run()



from typing import List
from typing import Set
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from netin import Graph


def update_name_homophily(data: Union[Graph, pd.DataFrame]) -> str:
    """
        Renames the object name to include the homophily values of the graph.
            :param data:
                :return:
                    """

    # valid = [Graph, pd.DataFrame]
    # test = [c for c in valid if isinstance(data, c)]
    # if len(test) == 0:
    #     raise TypeError(f"The data must be one of these types: {valid}")

    if pd.DataFrame == type(data):
        current = data.name
        h_MM = data.h_MM.unique()[0]
        h_mm = data.h_mm.unique()[0]
    else:
        current = data.get_model_name()
        h_MM = data.get_homophily_majority()
        h_mm = data.get_homophily_minority()

        to_add = r"h$_{MM}$=<hMM>, h$_{mm}$=<hmm>".replace("<hMM>", f"{h_MM}").replace("<hmm>", f"{h_mm}")
        return f"{current}\n{to_add}"

def plot_edge_type_counts(data: Union[Graph, list[Graph], set[Graph]], fn=None, **kwargs):
    """
    Plots the edge type counts of a single or multiple graphs
    :param data:
    :param kwargs:
    :return:
    """

    if type(data) not in [list, set, List, Set]:
        data = [data]

    w, h = len(data) * 3.2, 3.2  # default figure size (width, height)
    width = kwargs.pop('width_bar', 0.25)  # the width of the bars
    figsize = kwargs.pop('figsize', (w, h))  # figure size (width, height)
    loc = kwargs.pop('loc', 'upper right')  # position of legend
    ncols = kwargs.pop('nc_legend', 1)  # number of columns in legend

    fig, ax = plt.subplots(1, 1, figsize=figsize, layout='constrained')
    multiplier = 0

    x = None
    groups = None
    maxy = 0
    for g in data:
        etc = g.calculate_edge_type_counts()
        name = update_name_homophily(g)

        groups = list(etc.keys()) if groups is None else groups
        x = np.arange(len(groups)) if x is None else x
        y = [etc[i] for i in groups]
        maxy = max(max(y), maxy)

        offset = width * multiplier
        rects = ax.bar(x + offset, y, width, label=name)
        ax.bar_label(rects, padding=3)
        multiplier += 1

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_ylabel('Counts')
        ax.set_title("Counts of edge types");
        ax.set_xticks(x + width, groups)
        ax.legend(loc=loc, ncols=ncols)

        # set limits
        ax.set_ylim(0, maxy * 1.1)

    # save plot
    if fn is not None:
        validate_path(fn)
        fig.savefig(fname=fn, bbox_inches='tight', dpi=300)
        print(f'{fn} saved.')

    # Final
    plt.show()
    plt.close()

def validate_path(fn):
    import os
    path = os.path.dirname(fn)
    if path != '':
        if not os.path.exists(path):
            # Create a new directory because it does not exist
            os.makedirs(path)



from typing import Union, Set, Tuple

import numpy as np

from netin.generators.tc import TriadicClosure
from netin.utils import constants as const
from .pah import PAH


class PATCH(PAH, TriadicClosure):
    """Creates a new PATCH instance. An undirected graph with preferential attachment, homophily, and triadic closure.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    k: int
        minimum degree of nodes (minimum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    h_MM: float
        homophily (similarity) between majority nodes (minimum=0, maximum=1.)

    h_mm: float
        homophily (similarity) between minority nodes (minimum=0, maximum=1.)

    tc: float
        probability of a new edge to close a triad (minimum=0, maximum=1.)

    Notes
    -----
    The initialization is an undirected with n nodes and no edges.
    Then, everytime a node is selected as source, it gets connected to k target nodes.
    Target nodes are selected via preferential attachment (in-degree) [BarabasiAlbert1999]_,
    homophily (h_**; see :class:`netin.Homophily`) [Karimi2018]_,
    and triadic closure (see :class:`netin.TriadicClosure`) [HolmeKim2002]_.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, k: int, f_m: float, h_mm: float, h_MM: float, tc: float, seed: object = None):
        PAH.__init__(self, n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
        TriadicClosure.__init__(self, n=n, f_m=f_m, tc=tc, seed=seed)

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.PATCH_MODEL_NAME)

    def _validate_parameters(self):
        """
        Validates the parameters of the undirected.
        """
        PAH._validate_parameters(self)
        TriadicClosure._validate_parameters(self)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns a dictionary with the metadata of the PATCH graph.

        Returns
        -------
        dict
            the graph  metadata as a dictionary
        """
        obj1 = PAH.get_metadata_as_dict(self)
        obj2 = TriadicClosure.get_metadata_as_dict(self)
        obj1.update(obj2)
        return obj1

    ############################################################
    # Generation
    ############################################################

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                 special_targets: Union[None, object, iter] = None) -> Tuple[np.array, set[int]]:
        """
        Returns the probabilities of nodes to be selected as target nodes.

        Parameters
        ----------
        source: int
            source node id

        target_set: set
            set of target node ids

        special_targets: dict
            dictionary of special target node ids to be considered

        Returns
        -------
        tuple
            probabilities of nodes to be selected as target nodes, and set of target of nodes

        """
        return TriadicClosure.get_target_probabilities(self, source, target_set, special_targets)

    def get_target_probabilities_regular(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                         special_targets: Union[None, object, iter] = None) -> Tuple[
        np.ndarray, set[int]]:
        """
        Returns the probability of nodes to be selected as target nodes using the
        preferential attachment with homophily mechanism.

        Parameters
        ----------
        source: int
            source node id

        target_set: set
            set of target node ids

        special_targets: dict
            dictionary of special target node ids to be considered

        Returns
        -------
        tuple
            probabilities of nodes to be selected as target nodes, and set of target of nodes
        """
        return PAH.get_target_probabilities(self, source, target_set, special_targets)

    def get_special_targets(self, source: int) -> object:
        """
        Returns an empty dictionary (source node ids)

        Parameters
        ----------
        source : int
            Newly added node

        Returns
        -------
        Dict
            Empty dictionary
        """
        return TriadicClosure.get_special_targets(self, source)

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the (input) parameters of the graph.
        """
        PAH.info_params(self)
        TriadicClosure.info_params(self)

    def info_computed(self):
        """
        Shows the (computed) properties of the graph.
        """
        PAH.info_computed(self)
        TriadicClosure.info_computed(self)

    def infer_homophily_values(self) -> Tuple[float, float]:
        """
        Infers analytically the homophily values of the graph.

        Returns
        -------
        tuple
            homophily values of the graph (majority, minority)
        """
        h_MM = None
        h_mm = None
        return h_MM, h_mm

    def infer_triadic_closure(self) -> float:
        """
        Infers analytically the triadic closure value of the graph.

        Returns
        -------
        float
            triadic closure probability of the graph
        """
        tc = None
        return tc

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        return self.__class__(n=self.n,
                              k=self.k,
                              f_m=self.f_m,
                              tc=self.tc,
                              h_MM=self.h_MM,
                              h_mm=self.h_mm,
                              seed=self.seed)



from typing import Union, Set, Tuple

import numpy as np

from netin.utils import constants as const
from netin.utils import validator as val
from .graph import Graph


class Homophily(Graph):
    """Class to model homophily as a mechanism of edge formation given a source and a target node.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    h_MM: float
        homophily (similarity) between majority nodes (minimum=0, maximum=1.)

    h_mm: float
        homophily (similarity) between minority nodes (minimum=0, maximum=1.)

    Notes
    -----
    This class does not generate a graph.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, f_m: float, h_MM: float, h_mm: float, seed: object = None):
        Graph.__init__(self, n=n, f_m=f_m, seed=seed)
        self.h_MM = h_MM
        self.h_mm = h_mm
        self.mixing_matrix = None

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.H_MODEL_NAME)

    def _validate_parameters(self):
        """
        Validates the parameters of the graph.
        """
        Graph._validate_parameters(self)
        val.validate_float(self.h_MM, minimum=0., maximum=1.)
        val.validate_float(self.h_mm, minimum=0., maximum=1.)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns the metadata info (input parameters of the model) of the graph as a dictionary.

        Returns
        -------
        obj dict
            dictionary with the metadata info of the graph.
        """
        obj = Graph.get_metadata_as_dict(self)
        obj.update({
            'h_MM': self.h_MM,
            'h_mm': self.h_mm,
        })
        return obj

    ############################################################
    # Getters & Setters
    ############################################################

    def set_homophily_majority(self, h_MM: float):
        """
        Sets the homophily value between majority nodes.

        Parameters
        ----------
        h_MM: float
            homophily (similarity) between majority nodes (minimum=0, maximum=1.)
        """
        self.h_MM = h_MM

    def get_homophily_majority(self) -> float:
        """
        Returns the homophily value between majority nodes.

        Returns
        -------
        h_MM: float
            homophily (similarity) between majority nodes (minimum=0, maximum=1.)
        """
        return self.h_MM

    def set_homophily_minority(self, h_mm: float):
        """
        Sets the homophily value between minority nodes.

        Parameters
        ----------
        h_mm: float
            homophily (similarity) between minority nodes (minimum=0, maximum=1.)
        """
        self.h_mm = h_mm

    def get_homophily_minority(self) -> float:
        """
        Returns the homophily value between minority nodes.

        Returns
        -------
        h_mm: float
            homophily (similarity) between minority nodes (minimum=0, maximum=1.)
        """
        return self.h_mm

    def get_homophily_between_source_and_target(self, source: int, target: int) -> float:
        """
        Returns the homophily value between a source and a target node based on their class values.
        This homophily value is inferred from the mixing matrix.

        Parameters
        ----------
        source: int
            Source node id

        target: int
            Target node id

        Returns
        -------
        h: float
            homophily (similarity) between source and target nodes (minimum=0, maximum=1.)
        """
        return self.mixing_matrix[self.node_class_values[source], self.node_class_values[target]]

    ############################################################
    # Generation
    ############################################################

    def _initialize(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the model.

        Parameters
        ----------
        class_attribute: str
            name of the attribute that represents the class

        class_values: list
            values of the class attribute

        class_labels: list
            labels of the class attribute mapping the class_values.
        """
        Graph._initialize(self, class_attribute, class_values, class_labels)
        self.h_MM = val.calibrate_null_probabilities(self.h_MM)
        self.h_mm = val.calibrate_null_probabilities(self.h_mm)
        self.mixing_matrix = np.array([[self.h_MM, 1 - self.h_MM], [1 - self.h_mm, self.h_mm]])

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int], np.array],
                                 special_targets: Union[None, object, iter] = None) -> tuple[np.array, set[int]]:
        """
        Returns the probabilities of selecting a target node from a set of nodes based on homophily.
        Homophily is inferred from the mixing matrix.

        Parameters
        ----------
        source: int
            source node

        target_set: set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        tuple[np.array, set[int]]
            probabilities of selecting a target node from a set of nodes, and the set of target nodes`
        """
        probs = np.array([self.get_homophily_between_source_and_target(source, target) for target in target_set])
        probs /= probs.sum()
        return probs, target_set

    def get_target(self, source: Union[None, int], targets: Union[None, Set[int]],
                   special_targets: Union[None, object, iter]) -> int:
        """
        Picks a random target node based on the homophily dynamic.

        Parameters
        ----------
        source: int
            Newly added node

        targets: Set[int]
            Potential target nodes in the graph

        special_targets: object
            Special target nodes in the graph

        Returns
        -------
            int: Target node that an edge should be added to from `source`
        """
        # Collect probabilities to connect to each node in target_list
        target_set = self.get_potential_nodes_to_connect(source, targets)
        probs = self.get_target_probabilities(source, target_set, special_targets)
        return np.random.choice(a=target_set, size=1, replace=False, p=probs)[0]

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the parameters of the model.
        """
        print('h_MM: {}'.format(self.h_MM))
        print('h_mm: {}'.format(self.h_mm))
        print('mixing matrix: \n{}'.format(self.mixing_matrix))

    def info_computed(self):
        """
        Shows the computed properties of the graph.
        """
        inferred_h_MM, inferred_h_mm = self.infer_homophily_values()
        print("- Empirical homophily within majority: {}".format(inferred_h_MM))
        print("- Empirical homophily within minority: {}".format(inferred_h_mm))

    def infer_homophily_values(self) -> Tuple[float, float]:
        """
        Infers analytically the homophily values for the majority and minority classes.

        Returns
        -------
        h_MM: float
            homophily within majority group

        h_mm: float
            homophily within minority group
        """

        e = self.calculate_edge_type_counts()
        if self.is_directed():
            h_MM = e['MM'] / (e['MM'] + e['Mm'])
            h_mm = e['mm'] / (e['mm'] + e['mM'])
        else:
            h_MM = e['MM'] / (e['MM'] + e['Mm'] + e['mM'])
            h_mm = e['mm'] / (e['mm'] + e['mM'] + e['Mm'])

        return h_MM, h_mm



from typing import Union, Set

import numpy as np

from netin.generators.directed import DiGraph
from netin.utils import constants as const


class DPA(DiGraph):
    """Creates a new DPA instance. A directed graph with preferential attachment.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    d: float
        edge density (minimum=0, maximum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    plo_M: float
        activity (out-degree power law exponent) majority group (minimum=1)

    plo_m: float
        activity (out-degree power law exponent) minority group (minimum=1)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is a directed with n nodes and no edges.
    Source nodes are selected based on their activity given by plo_M (if majority) or plo_m (if minority) [Espin-Noboa2022]_.
    Target nodes are selected via preferential attachment [BarabasiAlbert1999]_.

    References
    ----------

    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, d: float, f_m: float, plo_M: float, plo_m: float, seed: object = None):
        super().__init__(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, seed=seed)

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.DPA_MODEL_NAME)

    ############################################################
    # Generation
    ############################################################

    def get_in_degree(self, n: int) -> int:
        """
        Returns the in-degree of node `n`.
        This in-degree is not calculated, it is taken from the object `in_degrees` that is populated while
        generating the graph.

        Parameters
        ----------
        n: int
            node id

        Returns
        -------
        int
            in-degree of node `n`
        """
        return self.in_degrees[n]

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int], np.array],
                                 special_targets: Union[None, object, iter] = None) -> np.array:
        """
        Returns the probabilities for each target node in `target_set` to be selected as target node
        given source node `source`.

        Parameters
        ----------
        source: int
            source node id

        target_set: Set[int]
            set of target node ids

        special_targets: object
            special targets

        Returns
        -------
        np.array
            array of probabilities for each target node.
        """
        probs = np.array([self.get_in_degree(n) + const.EPSILON for n in target_set])
        probs /= probs.sum()
        return probs

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        return self.__class__(n=self.n,
                              d=self.d,
                              f_m=self.f_m,
                              plo_M=self.plo_M,
                              plo_m=self.plo_m,
                              seed=self.seed)
